/*
(the) SquirrelBox compiler v1.0.
Copyright (C) 2005 Dan Cook

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

class Result { // Resulting value throughout an expression
	public Type type;    // dataType of result
	public String value; // numeric value or var address
	public boolean neg;  // negate the result before using
	public boolean not;  // invert the result before using
	public boolean address; // take vars by address or not

	public Result(Type t, String v, boolean ng, boolean nt, boolean a) {
		type = t; value = v; neg = ng; not = nt; address = a;
	}

	public Result(int s, String v, boolean ng, boolean nt, boolean a) {
		type = new Type(s, 0); value = v; neg = ng; not = nt; address = a;
	}

	public int dim() { return type.dim; }
	public int bytes() { return type.bytes; }
	public int size() { return type.size(); }

	public boolean numeric() {
		try {
			Integer.parseInt(value);
			return true;
		}
		catch(NumberFormatException n){}
		return false;
	}

	private boolean register(char ch) {
		return ch == 'H' || ch == 'L' || (ch >= 'A' && ch <= 'F');
	}

	public boolean register() {
		if(value.length()>2) return false;
		if(value.length()>1) {
			if(!register(value.charAt(1))) return false;
		} return register(value.charAt(0));
	}

	private void ranOutOfRegisters(HLz80compiler c, String b, boolean[] free) {
		String text =
			"Ran out of "+b+"-bit Registers to use :-(\nregisers in use: ";
		String reg = "ABCDEFHL";
		for(int i=0; i<8; i++) if(!free[i]) text += " "+reg.charAt(i);
		c.Error(text);
	}

	private void setSize() {
		try {
			if(Math.abs(Integer.parseInt(value)) < 256) type.bytes = 1;
			else type.bytes = 2;
		} catch(NumberFormatException n) {
			if(type.bytes != 2) type.bytes = 1;
		}
	}

	public boolean[] loadRegister(HLz80compiler c, boolean[] free) {
		if(register()) return free;
		if(type.dim > 0) c.Error(
			"cannot load array into register\n(value = "+toString()+')'
		);
		setSize();
		String ahl;
		String reg = null;
		if(type.bytes == 2) {
			if(free[6] && free[7]) {
				reg = "HL"; free[6] = false; free[7] = false;
			} else if(free[3] && free[4]) { // DE is better for EX DE,HL
				reg = "DE"; free[3] = false; free[4] = false;
			} else if(free[1] && free[2]) {
				reg = "BC"; free[1] = false; free[2] = false;
			} else if(free[0] && free[5]) {
				reg = "AF"; free[0] = false; free[5] = false;
			} else ranOutOfRegisters(c,"16",free);
			ahl = "HL";
		} else {
			if(free[0]) { reg = "A"; free[0] = false; }
			else if(free[1]) { reg = "B"; free[1] = false; }
			else if(free[2]) { reg = "C"; free[2] = false; }
			else if(free[3]) { reg = "D"; free[3] = false; }
			else if(free[4]) { reg = "E"; free[4] = false; }
			else if(free[5]) { reg = "F"; free[5] = false; }
			else if(free[6]) { reg = "H"; free[6] = false; }
			else if(free[7]) { reg = "L"; free[7] = false; }
			else ranOutOfRegisters(c, "8", free);
			ahl = "A";
		}
		if(!address && !numeric() && !reg.equals(ahl)) {
			if(ahl.equals("HL") && reg.equals("DE")) {
				c.asmOutln(" EX DE,HL");
				c.asmOutln(" LD HL,("+value+')');
				c.asmOutln(" EX DE,HL");
			} else {
				char temp = 0;
				if(free[0]) temp = 'A';
				else if(free[1]) temp = 'B';
				else if(free[2]) temp = 'C';
				else if(free[3]) temp = 'D';
				else if(free[4]) temp = 'E';
				else if(free[5]) temp = 'F';
				else if(free[6]) temp = 'H';
				else if(free[7]) temp = 'L';
				else ranOutOfRegisters(c, "8", free);
				if(ahl.equals("A")) {
					c.asmOutln(" LD "+temp+",A");
					c.asmOutln(" LD A,("+value+')');
					c.asmOutln(" LD "+reg+",A");
					c.asmOutln(" LD A,"+temp);
				} else {
					char temp2 = 0;
					if(free[0]) temp2 = 'A';
					else if(free[1]) temp2 = 'B';
					else if(free[2]) temp2 = 'C';
					else if(free[3]) temp2 = 'D';
					else if(free[4]) temp2 = 'E';
					else if(free[5]) temp2 = 'F';
					else if(free[6]) temp2 = 'H';
					else if(free[7]) temp2 = 'L';
					else ranOutOfRegisters(c, "8", free);
					c.asmOutln(" LD "+temp+",H");
					c.asmOutln(" LD "+temp2+",L");
					c.asmOutln(" LD HL,("+value+')');
					c.asmOutln(" LD "+reg.charAt(0)+",H");
					c.asmOutln(" LD "+reg.charAt(1)+",L");
					c.asmOutln(" LD H,"+temp);
					c.asmOutln(" LD L,"+temp2);
				}
			}
		} else c.asmOutln(" LD "+reg+","+asmValue());
		address = false; value = reg;
		return free;
	}

	public String unused(HLz80compiler c, boolean[] free, boolean word) {
		setSize();
		if(word) {
			if(free[6] && free[7]) return "HL";
			if(free[1] && free[2]) return "BC";
			if(free[3] && free[4]) return "DE";
			if(free[0] && free[5]) return "AF";
			ranOutOfRegisters(c,"16",free);
		}
		if(free[0]) return "A";
		if(free[1]) return "B";
		if(free[2]) return "C";
		if(free[3]) return "D";
		if(free[4]) return "E";
		if(free[5]) return "F";
		if(free[6]) return "H";
		if(free[7]) return "L";
		ranOutOfRegisters(c,"8",free);
		return null;
	}

	public String low() {
		try {
			return Integer.toString(Integer.parseInt(value)%256);
		} catch(NumberFormatException n){}
		if(register()) {
			if(value.length() == 2)
				return value.substring(1);
			return value;
		}
		if(address) return value+"%256";
		return "("+value+')';
	}

	public String high() {
		try {
			return Integer.toString(Integer.parseInt(value)/256);
		} catch(NumberFormatException n){}
		if(register()) {
			if(value.length() == 2)
				return value.substring(0,1);
			return "0";
		}
		if(address) {
			if(type.bytes == 2)
				return value+"/256";
			return value+"%256";
		}
		if(type.bytes == 2)
			return "("+value+"+1)";
		return "0";
	}

	public String getByte(int high) {
		if(high > 0) return high();
		return low();
	}

	public String toString() {
		String str = "";
		if(neg) str = "-";
		if(not) str += '~';
		return str + value;
	}

	public String asmValue() {
		if(!register() && !numeric() && !address) return "("+value+')';
		return value;
	}

	public static boolean[] freeList() {
		return new boolean[]{true,true,true,true,true,true,true,true};
	}

	public static boolean[] freeList(String use) {
		boolean[] free = freeList();
		for(int i=0; i<use.length(); i++) {
			switch(use.charAt(i)) {
			  case 'A': free[0] = false;
			  case 'B': free[1] = false;
			  case 'C': free[2] = false;
			  case 'D': free[3] = false;
			  case 'E': free[4] = false;
			  case 'F': free[5] = false;
			  case 'H': free[6] = false;
			  case 'L': free[7] = false;
			}
		} return free;
	}
}

class Context { // Code context within [nested] control-structures
	public int blocks;
	public String trail;
	public Context previous;
	public variable[] locals;
	public int cont, brk, next, exit;

	public Context() {
		blocks = 0; trail = "."; previous = null;
		locals = new variable[0]; cont = -1; brk = -1; next = -1; exit = -1;
	}

	private Context(String t, Context prev, int c, int b, int n, int e) {
		blocks = 0; trail = t; previous = prev;
		locals = new variable[0]; cont = c; brk = b; next = n; exit = e;
	}

	public Context passIf() {
		return new Context(trail+(blocks++)+'.', this, cont, brk, next, exit);
	}

	public Context passLoop(int c, int b) {
		return new Context(trail+(blocks++)+'.', this, c, b, next, exit);
	}

	public Context passSwitch(int n, int e) {
		return new Context(trail+(blocks++)+'.', this, cont, brk, n, e);
	}

	public int offset() {
		if(previous == null) return 0;
		return previous.offset()+previous.locals.length;
	}

	public int length() {
		return offset()+locals.length;
	}

	public int find(String var) {
		int index;
		if(previous != null) {
			index = previous.find(var);
			if(index >= 0) return index;
		}
		index = utils.find(locals, var);
		if(index >= 0) return offset()+index;
		return -1;
	}

	public variable varAt(int index) {
		if(previous == null) {
			if(index >= locals.length) return null;
			return locals[index];
		}
		variable var = previous.varAt(index);
		if(var != null) return var;
		return locals[index-offset()];
	}

	public String nameAt(int index) {
		if(previous == null) {
			if(index >= locals.length) return null;
			return trail+locals[index].name;
		}
		String name = previous.nameAt(index);
		if(name != null) return name;
		return trail+locals[index-offset()].name;
	}

	public void remove(int index) {
		if(index < offset()) {
			previous.remove(index); return;
		}
		index -= offset();
		variable[] temp = new variable[locals.length-1];
		for(int i=0; i<index; i++) temp[i] = locals[i];
		for(int i=index; i<temp.length; i++) temp[i] = locals[i+1];
		locals = temp;
	}
}

public class HLz80compiler extends HLz80gui {
	public int Labels;
	public boolean noReturn;
	public boolean usecphlde;
	public String[] asmincs;
	public variable[] globals;
	public String[] inits;
	public function[] funcs;
	public String[][] memory;

	public int NewLabel() {
		return Labels++;
	}

	public void comment(String com) {
		if(com == null) asmOutln();
		else {
			String text = asmOutput[current].getText();
			int fix = text.lastIndexOf('\n');
			if(fix < 0) fix = text.length()+1;
			else fix = text.length() - fix;
			if(fix > 16) fix = 9 + fix%8;
			asmOutln("                 ; ".substring(fix)+com);
		}
	}

	public void PostLabel(int label, String com) {
		asmOut("Label"+Integer.toString(label)+':');
		comment(com);
	}

	public int PostNewLabel(String comment) {
		PostLabel(Labels, comment);
		return Labels++;
	}

	public void jump(int label, String com) {
		asmOut(" JP Label"+label); comment(com);
	}

	public int[] typeOf(Context context, String token) { // {type, index}
		int index = utils.find(funcs, token);
		if(index >= 0) return new int[]{0,index};
		index = utils.find(globals, token);
		if(index >= 0) return new int[]{1,index};
		index = context.find(token);
		if(index >= 0) return new int[]{2,index};
		if(token.equals("this") && func != null) {
			if(func.ret.type.bytes > 0) return new int[]{3};
		} return new int[]{-1};
	}

	public boolean isDefined(Context context, String token) {
		return typeOf(context, token)[0] != -1;
	}

	public int endOfText() {
		return asmOutput[current].getText().length();
	}

	public void insert(int place, String text) {
		String asm = asmOutput[current].getText();
		if(!loading) asmOutput[current].setText(
			asm.substring(0,place)+text+asm.substring(place)
		);
	}

	public void use(Context context, String item) {
		if(loading) func.use(item);
		else func.unUse(context, item);
	}

	public void clearUnused(Context context) {
		if(!loading) func.clearUnused(context);
	}

	public Result getValue(Context context, int size, boolean ng, boolean nt) {
		int num = in.getSingleNum();
		if(num > 65535 || (num > 255 && size == 1))
			Warning("Value is too big and will be rounded: "+num);
		if(ng) num = ~num;
		if(nt) num = -num;
		if(size == 1) num %= 256;
		else num %= 65536;
		return new Result(size, Integer.toString(num), false, false, false);
	}

	public void parseFunctionCall(Context context, function f, boolean[] free) {
		// assume that function name and "(" are already parsed
		Result result;
		String reg = null;
		for(int i=0; i<f.args.length; i++) {
			if(i > 0) in.MatchChar(',');
			result = parseExpression(context, f.args[i].type, free, false);
			if(
				result.value.equals("A") || result.value.equals("BC") ||
				result.value.equals("DE") || result.value.equals("AF")
			) asmOutln(" LD ("+f.args[i].address+"),"+result.asmValue());
			else if(result.type.bytes == 2) {
				if(!free[6] || !free[7]) {
					reg = result.unused(this, free, false);
					if(reg.equals("DE")) asmOutln(" EX DE,HL");
					else {
						asmOutln(" LD "+reg.charAt(0)+",H");
						asmOutln(" LD "+reg.charAt(1)+",L");
					}
				} asmOutln(" LD HL,"+result.asmValue());
				asmOutln(" LD ("+f.args[i].address+"),A");
				if(!free[6] || !free[7]) {
					if(reg.equals("DE")) asmOutln(" EX DE,HL");
					else {
						asmOutln(" LD H,"+reg.charAt(0));
						asmOutln(" LD L,"+reg.charAt(1));
					}
				}
			} else {
				if(!free[0]) {
					reg = result.unused(this, free, false);
					asmOutln(" LD "+reg+",A");
				} asmOutln(" LD A,"+result.asmValue());
				asmOutln(" LD ("+f.args[i].address+"),A");
				if(!free[0]) asmOutln(" LD A,"+reg);
			}
		}
		asmOutln(" CALL func_"+f.ret.name);
		use(context, f.ret.name);
		in.MatchChar(')');
	}

	private Result anyDimFactor(
		Context context, Type type, boolean[] free, boolean nfirst
	) {
		byte which = 0;
		boolean neg = in.getCharIf('-');
		boolean not = in.getCharIf('~');
		if(in.getCharIf('(')) {
			if(in.getIdent().equals("LSB")) which = 1;
			else if(in.Matches("MSB")) which = 2;
			else Error("No parenthesis allowed in expressions");
			in.MatchChar(')');
		} boolean address = in.getCharIf('&');
		in.checkForAsm(this, context);
		if(in.numeric()) {
			if(which > 0) Error("You can only take the LSB or MSB of a word");
			if(address) Error(
				"You cannot take the address of a number", "&"+in.getSingleNum()
			); return getValue(context, type.bytes, neg, not);
		}
		int[] typeof = typeOf(context, in.getIdent());
		variable var = null;
		switch(typeof[0]) {
		  case 0:
			if(which > 0 && funcs[typeof[1]].ret.type.bytes != 2)
				Error("You can only take the LSB or MSB of a word");
			if(in.getCharIf('(')) {
				if(address) Error(
					"You cannot take the address of a function call",
					"&"+funcs[typeof[1]].ret.name+'('
				);
				if(nfirst) Error(
					"Function calls must come first in expressions",
					"call to function "+funcs[typeof[1]].ret.name
				);
				use(context, funcs[typeof[1]].ret.name);
				parseFunctionCall(context, funcs[typeof[1]], free);
				in.checkForAsm(this, context);
				if(which == 0) return new Result(
					funcs[typeof[1]].ret.type,
					funcs[typeof[1]].ret.address, neg, not, false
				);
				if(which == 1) return new Result(
					new Type(1, funcs[typeof[1]].ret.type.dim),
					funcs[typeof[1]].ret.address, neg, not, false
				);
				return new Result(
					new Type(1, funcs[typeof[1]].ret.type.dim),
					funcs[typeof[1]].ret.address+"+1", neg, not, false
				);
			}
			if(address) var = new variable(
				new Type(2,0), null, "func_"+funcs[typeof[1]].ret.name
			); else Error("Illegal use of function "+in.token+" as a value");
			break;
		  case 1:
			var = globals[typeof[1]];
			use(context, var.name);
			break;
		  case 2:
			var = context.varAt(typeof[1]);
			use(context, context.nameAt(typeof[1]));
			break;
		  case 3: var = func.ret; break;
		  default: Error("Invalid item in expression", in.token);
		} in.checkForAsm(this, context);
		if(which == 0) {
			if(address) return new Result(
				new Type(2, var.type.dim), var.address, neg, not, address
			); return new Result(var.type, var.address, neg, not, address);
		}
		if(which == 1) return new Result(
			new Type(1, var.type.dim), var.address, neg, not, address
		); return new Result(
			new Type(1, var.type.dim), var.address+"+1", neg, not, address
		);
	}

	public Result parseFactor(
		Context context, Type type, boolean[] free, boolean nfirst
	) {
		Result factor = anyDimFactor(context, type, free, nfirst);
		if( (type.dim != factor.type.dim && type.dim > 0) ||
			(type.bytes < factor.type.bytes && type.bytes > 0)
		) Error(type+" expected", factor.type.toString());
		if(in.getCharIf('[')) {
			if(factor.type.dim < 1) Error("You can only index an array");
			if(type.dim > 0) Error("Array expected, array index found");
			Result result = parseExpression(
				context, new Type(type.bytes, type.dim-1), free, nfirst
			); in.MatchChar(']');
			if(factor.numeric() && result.numeric())
				factor.value += "+"+result.value;
			else {
				String reg = null;
				free = factor.loadRegister(this, free);
				boolean check = result.register() && result.type.bytes == 2;
				if(factor.equals("A")) {
					if(result.neg) {
						if(result.not) asmOutln(
							" SCF\n ADC "+factor.value+','+result.asmValue()
						); else asmOutln(
							" SUB "+factor.value+','+result.asmValue()
						);
					} else if(result.not) asmOutln(
						" SUB "+factor.value+','+result.asmValue()+
						"\n INC "+factor.value
					); else asmOutln(
						" ADD "+factor.value+','+result.asmValue()
					);
				} else if(factor.type.bytes == 2 && check) {
					if(!factor.value.equals("HL")) {
						if(!free[6] || !free[7]) {
							reg = result.unused(this, free, true);
							if(reg.equals("DE")) asmOutln(" EX DE,HL");
							else {
								asmOutln(" LD "+reg.charAt(0)+",H");
								asmOutln(" LD "+reg.charAt(1)+",L");
							}
						}
						if(factor.value.equals("DE")) asmOutln(" EX DE,HL");
						else {
							asmOutln(" LD H,"+factor.value.charAt(0));
							asmOutln(" LD L,"+factor.value.charAt(1));
						}
					}
					if(result.neg) {
						asmOutln(" SCF");
						if(result.not) asmOutln(" ADC HL,"+result.value);
						else asmOutln(" CCF\n SBC HL,"+result.value);
					} else if(result.not) asmOutln(
						" SCF\n SBC HL,"+result.value
					); else asmOutln(" ADD HL,"+result.value);
					if(!factor.value.equals("HL")) {
						if(factor.value.equals("DE")) asmOutln(" EX DE,HL");
						else {
							asmOutln(" LD "+factor.value.charAt(0)+",H");
							asmOutln(" LD "+factor.value.charAt(1)+",L");
						}
						if(!free[6] || !free[7]) {
							if(reg.equals("DE")) asmOutln(" EX DE,HL");
							else {
								asmOutln(" LD H,"+reg.charAt(0));
								asmOutln(" LD L,"+reg.charAt(1));
							}
						}
					}
				} else {
					if(!free[0]) {
						reg = result.unused(this, free, false);
						asmOutln(" LD "+reg+",A");
					}
					for(int i=factor.type.bytes; i>0; i--) {
						check = i == 0 && factor.type.bytes == 2;
						asmOutln(" LD A,"+factor.value.charAt(i-1));
						if(result.neg) {
							if(result.not) {
								if(!check) asmOutln(" SCF");
								asmOutln(" ADC A,"+result.getByte(i));
							} else {
								if(check) asmOutln(" SBC A,"+result.getByte(i));
								else asmOutln(" SUB A,"+result.getByte(i));
							}
						} else if(result.not) {
							if(check) asmOutln(" SBC A,"+result.getByte(i));
							else asmOutln(
								" SUB A,"+result.getByte(i)+"\n INC A"
							);
						} else {
							if(check) asmOutln(" ADC A,"+result.getByte(i));
							else asmOutln(" ADD A,"+result.getByte(i));
						} asmOutln(" LD "+result.getByte(i)+",A");
					} if(!free[0]) asmOutln(" LD A,"+reg);
				}
				if(!factor.address) { //load value at the address or the address
					if(factor.type.bytes == 2) {
						if(!factor.value.equals("HL")) {
							if(!free[6] && !free[7]) {
								reg = factor.unused(this, free, true);
								if(reg.equals("DE")) asmOutln(" EX DE,HL");
								else {
									asmOutln(" LD "+reg.charAt(0)+",H");
									asmOutln(" LD "+reg.charAt(1)+",L");
								}
							}
						} asmOutln(" LD HL,("+factor.value+')');
						if(!factor.value.equals("HL")) {
							if(factor.value.equals("DE")) asmOutln(" EX DE,HL");
							else {
								asmOutln("LD "+factor.value.charAt(0)+",H");
								asmOutln("LD "+factor.value.charAt(1)+",L");
							}
							if(!free[6] && !free[7]) {
								if(reg.equals("DE")) asmOutln(" EX DE,HL");
								else {
									asmOutln(" LD H,"+reg.charAt(0));
									asmOutln(" LD L,"+reg.charAt(1));
								}
							}
						}
					} else {
						String reg2 = factor.unused(this, free, true);
						asmOutln(" LD "+reg2.charAt(0)+",0");
						asmOutln(" LD "+reg2.charAt(1)+','+factor.value);
						if(!free[0] && !factor.equals("A")) {
							reg = factor.unused(this, free, false);
							asmOutln(" LD "+reg+",A");
						} asmOutln(" LD A,("+reg2+')');
						if(!free[0] && !factor.equals("A"))
							asmOutln(" LD "+reg+",A");
					}
					asmOutln(" LD "+factor.value+",("+factor.value+')');
				}
				if(factor.neg || factor.not) {
					if(!factor.equals("A")) {
						if(!free[0]) {
							reg = factor.unused(this, free, false);
							asmOutln(" LD "+reg+",A");
						} asmOutln(" LD A,"+factor.value);
					}
					if(factor.neg) {
						if(factor.not) asmOutln(" INC "+reg);
						else asmOutln(" NEG");
					} else if(factor.not) asmOutln(" CPL");
					if(!factor.equals("A")) {
						asmOutln(" LD "+factor.value+",A");
						if(!free[0]) asmOutln(" LD A,"+reg);
					}
				}
			} factor.not = false; factor.neg = false;
		} else if(type.bytes > 0 && type.dim < 1 && factor.type.dim > 0)
			Error("Array found where a value was expected");
		if((factor.neg || factor.not) && type.dim > 0)
			Error("Syntax Error: (attempted to negate or invert an array)");
		in.checkForAsm(this, context);
		return factor;
	}

	public Result parseTerm(
		Context context, Type type, boolean[] free, boolean nfirst
	) {
		char op;
		Result factor;
		boolean check;
		boolean[] copy;
		String save = null;
		String ins, reg = null;
		Result result = parseFactor(context, type, free, nfirst);
		in.mark();
		do {
			in.reset();
			copy = free;
			if(in.eof()) {
				in.unMark(); break;
			}
			op = in.getChar();
			if(op == '&') ins = " AND ";
			else if(op == '|') ins = " OR ";
			else if(op == '^') ins = " XOR ";
			else { in.pushBack(); break; }
			check = result.numeric();
			if(check) {
				reg = result.value;
				save = asmOutput[current].getText();
			} copy = result.loadRegister(this, copy);
			factor = parseFactor(context, type, copy, true);
			if(check && factor.numeric()) {
				asmOutput[current].setText(save);
				int value = Integer.parseInt(reg);
				if(op == '&') value &= Integer.parseInt(factor.value);
				else if(op == '|') value |= Integer.parseInt(factor.value);
				else value ^= Integer.parseInt(factor.value);
				result.value = Integer.toString(value);
			}
			else if(result.value.equals("A")) asmOutln(ins+factor.asmValue());
			else {
				if(!copy[0]) {
					reg = factor.unused(this, copy, false);
					asmOutln(" LD "+reg+",A");
				}
				for(int i=0; i<result.value.length(); i++) {
					asmOutln(" LD A,"+result.value.charAt(i));
					asmOutln(ins+factor.getByte(i));
					asmOutln(" LD "+result.value.charAt(i)+",A");
				} if(!copy[0]) asmOutln(" LD A,"+reg);
			}
		} while(true);
		return result;
	}

	public Result parseExpression(
		Context context, Type type, boolean[] free, boolean nfirst
	) {
		char op;
		Result term;
		boolean check;
		boolean[] copy;
		String reg = null;
		String save = null;
		String inc, dec, add;
		Result result = parseTerm(context, type, free, nfirst);
		in.mark();
		do {
			in.reset();
			copy = free;
			if(in.eof()) {
				in.unMark(); break;
			}
			op = in.getChar();
			if(op == '+') { inc = " INC "; dec = " DEC "; add = " ADD "; }
			else if(op == '-') { inc = " DEC "; dec = " INC "; add = " SUB "; }
			else { in.pushBack(); break; }
			check = result.numeric();
			if(check) {
				reg = result.value;
				save = asmOutput[current].getText();
			} copy = result.loadRegister(this, copy);
			term = parseTerm(context, type, copy, true);
			if(check && term.numeric()) {
				asmOutput[current].setText(save);
				int value = Integer.parseInt(reg);
				if(op == '+') value += Integer.parseInt(term.value);
				else value -= Integer.parseInt(term.value);
				result.value = Integer.toString(value);
			}
			else if(term.value.equals("1")) asmOutln(inc+result.value);
			else if(
				term.value.equals("65535") || (
					term.type.bytes == 1 && term.value.equals("255")
				)
			) asmOutln(dec+result.value);
			else if(result.value.equals("A")) asmOutln(add+"A,"+term.asmValue());
			else if(result.type.bytes == 2) {
				if(!result.value.equals("HL")) {
					if(!copy[6] || !copy[7]) {
						reg = term.unused(this, copy, true);
						if(reg.equals("DE")) asmOutln(" EX DE,HL");
						else {
							asmOutln(" LD "+reg.charAt(0)+",H");
							asmOutln(" LD "+reg.charAt(1)+",L");
						}
					} asmOutln(" LD H,"+result.value.charAt(0));
					asmOutln(" LD L,"+result.value.charAt(1));
				} asmOutln(add+"HL,"+term.asmValue());
				if(!result.value.equals("HL")) {
					asmOutln(" LD "+result.value.charAt(0)+",H");
					asmOutln(" LD "+result.value.charAt(1)+",L");
					if(!copy[6] || !copy[7]) {
						if(reg.equals("DE")) asmOutln(" EX DE,HL");
						else {
							asmOutln(" LD H,"+reg.charAt(0));
							asmOutln(" LD L,"+reg.charAt(1));
						}
					}
				}
			} else {
				if(!copy[0]) {
					reg = term.unused(this, copy, false);
					asmOutln(" LD "+reg+",A");
				} asmOutln(" LD A,"+result.value);
				asmOutln(add+"A,"+term.asmValue());
				asmOutln(" LD "+result.value+",A");
				if(!copy[0]) asmOutln(" LD A,"+reg);
			}
		} while(true);
		return result;
	}

	public void parseCondition(Context context, String jpLbl, boolean swap) {
		String flag;
		boolean not = false;
		boolean[] free = null;
		if(in.next() == 'n') {
			in.mark();
			if(in.getIdent().equals("not")) {
				if(in.getCharIf('(')) {
					not = true; in.unMark();
				} else in.pushBack();
			} else in.pushBack();
		} else if(in.next() == '!') {
			in.mark(); in.getChar();
			if(in.getCharIf('(')) {
				in.unMark(); not = true;
			} else in.pushBack();
		}
		if(in.next() >= 'C' && in.next() <= 'Z') {
			in.mark(); flag = null;
			if(in.getIdent().equals("CFLAG")) flag = "C";
			else if(in.Matches("HFLAG")) flag = "H";
			else if(in.Matches("NCFLAG")) flag = "NC";
			else if(in.Matches("NHFLAG")) flag = "NH";
			else if(in.Matches("NNFLAG")) flag = "NN";
			else if(in.Matches("NFLAG")) flag = "N";
			else if(in.Matches("NSFLAG")) flag = "NS";
			else if(in.Matches("NZFLAG")) flag = "NZ";
			else if(in.Matches("PEFLAG") || in.Matches("POFLAG")) {
				if(in.Matches("PEFLAG")^swap) flag = "PO";
				else flag = "PO";
			} else if(in.Matches("SFLAG")) flag = "S";
			else if(in.Matches("ZFLAG")) flag = "Z";
			else if(in.Matches("TRUE") || in.Matches("FALSE")) {
				if(in.Matches("FALSE")^swap) asmOutln(" JP "+jpLbl);
				return;
			}
			if(flag == null)
				in.pushBack();
			else {
				in.unMark();
				if((flag.charAt(0) == 'N')^swap)
					flag = flag.substring(1);
				else if(flag.charAt(0) != 'P')
					flag += 'N';
				asmOutln(" JP "+flag+','+jpLbl);
				return;
			}
		}
		Result lhs = parseExpression(
			context, new Type(0,0), Result.freeList(), false
		);
		char op = in.get();
		switch(op) {
		  case '=':
			in.MatchChar('='); op = 0; break;   // ==
		  case '!':
			if(in.getCharIf('>')) op = 4;       // !<
			else if(in.getCharIf('<')) op = 2;  // !>
			else { in.MatchChar('='); op = 1; } // !=
			break;
		  case '>':
		  	if(in.getCharIf('=')) op = 2; // >=
		  	else if(swap) op = 5;         // >
		  	else op = 4; break;           // <=
		  case '<':
		  	if(in.getCharIf('=')) {
		  		if(swap) op = 4;  // <=
		  		else op = 5;      // >
		  	} else op = 3; break; // <
		  default: Error(
				"Relational operator expected (next char is \'"+op+"\')"
			);
		} Result rhs, temp;
		if(lhs.register()) rhs = parseExpression(
			context, lhs.type, Result.freeList(lhs.value), true
		);
		else rhs = parseExpression(context, lhs.type, Result.freeList(), false);
		if(not) in.MatchChar(')');
		if(lhs.numeric() && rhs.numeric()) {
			int a = Integer.parseInt(lhs.value);
			int b = Integer.parseInt(rhs.value);
			if( (   (op==0 && a==b) || (op==3 && a<b) || (op==5 && a>b) ||
					(op==1 && a!=b) || (op==4 && a<=b) || (op==2 && a>=b)
				) == swap
			) asmOutln(" JP "+jpLbl);
		}
		if(
			rhs.value.equals("HL") ||
			rhs.value.equals("A") && lhs.type.bytes != 2
		) {
			if(op > 1) {
				if(op < 4) op += 2;
				else op -= 2;
			} temp = rhs;
			rhs = lhs;
			lhs = temp;
		} int mid;
		if(op < 2) flag = "Z"; else flag = "C";
		if(swap^(op==0 || op==3)) flag = "N"+flag;
		free = lhs.loadRegister(this, Result.freeList(rhs.value));
		if(lhs.type.bytes == 2 || rhs.type.bytes == 2) {
			rhs.loadRegister(this, free);
			asmOutln(";THE FOLLOWING BETTER SAY HL: "+lhs.value);
			asmOutln(";THE FOLLOWING BETTER SAY DE: "+rhs.value);
			asmOutln(";IF NOT, REPORT IT AS AN ERROR");
			asmOutln("CALL cphlde");
		} else {
			asmOutln(";THE FOLLOWING BETTER SAY A: "+lhs.value);
			asmOutln(";IF NOT, REPORT IT AS AN ERROR");
			if(rhs.numeric() || rhs.register() || rhs.address)
				asmOutln(" CP "+rhs.value);
			else {
				String reg = null;
				if(!free[6] || !free[7]) {
					reg = rhs.unused(this, free, true);
					if(reg.equals("DE")) asmOutln(" EX DE,HL");
					else {
						asmOutln(" LD "+reg.charAt(0)+",H");
						asmOutln(" LD "+reg.charAt(1)+",L");
					}
				} asmOutln(" LD HL,"+rhs.value);
				asmOutln(" CP (HL)");
				if(!free[6] || !free[7]) {
					if(reg.equals("DE")) asmOutln(" EX DE,HL");
					else {
						asmOutln(" LD H,"+reg.charAt(0));
						asmOutln(" LD L,"+reg.charAt(1));
					}
				}
			}
		} if(op == 5) {
			mid = NewLabel();
			asmOutln(" JP C,Label"+mid);
			asmOutln(" JP NZ,"+jpLbl);
			PostLabel(mid,"mid-conditional-pass/fail");
		} if(op == 4) {
			asmOutln(" JP C,"+jpLbl);
			asmOutln(" JP Z,"+jpLbl);
		} else asmOutln(" JP "+flag+','+jpLbl);
	}

	public void parseGoto(Context context) {
		String label = in.getIdent();
		if(utils.find(func.labels, in.token) == -1) Error(
			"Invalid goto; no \""+in.token+
			"\" label in function "+func.ret.name
		); label = "Lbl_"+func.ret.name+'_'+label;
		boolean If = !in.eof();
		if(If) If = in.frontChar(in.next());
		if(If) {
			in.mark(); If = in.getIdent().equals("if");
			if(If) in.unMark(); else in.pushBack();
		}
		if(If) {
			in.MatchChar('(');
			parseCondition(context, label, true);
			in.MatchChar(')');
		} else asmOutln(" JP "+label);
		in.MatchChar(';');
	}

	public void parseIf(Context context) {
		in.MatchChar('(');
		comment("<IF>");
		int L1 = NewLabel();
		parseCondition(context, "Label"+L1, false);
		in.MatchChar(')');
		if(parseBlock(context.passIf(), true)) {
			int L2 = NewLabel();
			jump(L2,"skip else code");
			PostLabel(L1, "<ELSE>");
			parseBlock(context.passIf(), false);
			PostLabel(L2,"<End IF-ELSE>");
			return;
		} else PostLabel(L1,"<End IF>");
	}

	public void parseWhile(Context context, boolean until) {
		String which;
		if(until) which = "UNTIL";
		else which = "WHILE";
		in.MatchChar('(');
		int L1 = PostNewLabel("<"+which+'>');
		int L2 = NewLabel();
		parseCondition(context, "Label"+L2, until);
		in.MatchChar(')');
		parseBlock(context.passLoop(L1, L2), false);
		jump(L1, "continue "+which);
		PostLabel(L2, "<End "+which+'>');
	}

	public void parseDo(Context context) {
		String which = null;
		int L1 = PostNewLabel("<DO->");
		int top = endOfText()-1;
		int cond = NewLabel();
		int L2 = NewLabel();
		parseBlock(context.passLoop(cond, L2), false);
		if(in.getIdent().equals("until")) which = "UNTIL";
		else if(in.token.equals("while")) which = "WHILE";
		else Error(
			"Missing condition after do (while or until expected)", in.token
		);
		in.MatchChar('(');
		parseCondition(context, "Label"+L1, which.equals("WHILE"));
		in.MatchChar(')');
		insert(top, which);
		PostLabel(L2,"<End "+which);
	}

	public void parseLoop(Context context) {
		int L1 = PostNewLabel("<LOOP>");
		int L2 = NewLabel();
		parseBlock(context.passLoop(L1, L2), false);
		jump(L1,"continue LOOP");
		PostLabel(L2,"<End LOOP>");
	}

	public void parseFor(Context context) {
		comment("<FOR>");
		in.MatchChar('(');
		variable var = null;
		int[] typeof = typeOf(context, in.getIdent());
		switch(typeof[0]) {
		  case 0:
			Error("Variable expected in for-loop", "function "+in.token);
		  case 1:
			var = globals[typeof[1]];
			use(context, var.name);
			break;
		  case 2:
			var = context.varAt(typeof[1]);
			use(context, context.nameAt(typeof[1]));
			break;
		  case 3: var = func.ret; break;
		  default: Error(in.token+" is undefined", "for("+in.token);
		} in.MatchChar(',');
		if(var.type.dim > 0) Error(
			"The \"for\" variable cannot be an array", "for("+var.type+"..."
		);
		Result inc = parseExpression(
			context, var.type, Result.freeList(), false
		); in.MatchChar(',');
		inc.loadRegister(this, Result.freeList());
		asmOutln(" LD ("+var.address+"),"+inc.value);
		Result result = parseExpression(
			context, var.type, Result.freeList(), false
		);
		if(result.register()) Error(
			"The ending value for a for-loop\n" +
			"must be a variable or numeric value\n" +
			"(in assembly, result was \""+inc+"\")"
		);
		if(in.getCharIf(',')) {
			inc = parseExpression(context, var.type, Result.freeList(), false);
			if(inc.register()) Error(
				"The incremental value for a for-loop\n" +
				"must be a variable or numeric value\n" +
				"(in assembly, result was \""+inc+"\")"
			);
		} else inc = new Result(1, "1", false, false, false);
		in.MatchChar(')');
		int top = PostNewLabel("<FOR pass>");
		int end = NewLabel();
		parseBlock(context.passLoop(top, end), false);
		String reg = "A";
		if(var.type.bytes == 2) reg = "HL";
		asmOutln(" LD "+reg+",("+var.address+')');
		if(reg.equals("A")) asmOutln(" ADD A,"+inc.asmValue());
		else asmOutln(" LD BC,"+inc.asmValue()+"\n ADD HL,BC");
		asmOutln(" LD ("+var.address+"),"+reg);
		if(reg.equals("A")) asmOutln(" CP "+result.asmValue());
		else {
			asmOutln(" LD A,H\n CP "+result.high());
			asmOutln(" JP NC,Label"+top);
			asmOutln(" LD A,L\n CP "+result.low());
		}
		asmOutln(" JP NC,Label"+top);
		PostLabel(end,"<End FOR>");
	}

	public void parseSwitch(Context context) {
		comment("<SWITCH>");
		in.MatchChar('(');
		String name = null;
		variable var = null;
		int[] typeof = typeOf(context, in.getIdent());
		switch(typeof[0]) {
		  case 0:
			Error("Variable expected in switch", "function "+in.token);
		  case 1:
			var = globals[typeof[1]];
			use(context, var.name);
			break;
		  case 2:
			var = context.varAt(typeof[1]);
			use(context, context.nameAt(typeof[1]));
			break;
		  case 3: var = func.ret; break;
		  default: Error(in.token+" is undefined", "switch("+in.token);
		} in.MatchChar(')');
		Result result;
		boolean loop = false;
		boolean notdef = false;
		int next, pass, end = NewLabel();
		boolean bracket = in.getCharIf('{');
		if(!bracket) in.MatchIdent("begin");
		do {
			if(bracket) { if(in.getCharIf('}')) break; }
			in.getIdent();
			if(!bracket && in.Matches("end")) break;
			next = pass = NewLabel();
			notdef = in.Matches("case");
			if(notdef) {
				comment("<CASE>");
				in.MatchChar('(');
				do {
					result = parseExpression(
						context, new Type(0,0), Result.freeList(), false
					); result.loadRegister(this, Result.freeList());
					if(!result.type.equals(var.type)) Error(
						"Invalid case; "+result.type+" found where "+
						var.type+" was expected"
					);
					if(in.getCharIf(',')) {
						if(!loop) { next = NewLabel(); loop = true; }
					} else loop = false;
					asmOutln(" CP "+result.asmValue()+",("+var.address+')');
					if(loop) asmOutln(" JP Z,Label"+pass);
					else {
						asmOutln(" JP NZ,Label"+next);
						if(pass != next) PostLabel(pass, "<CASE pass>");
					}
				} while(loop);
				in.MatchChar(')');
			} else {
				if(!in.Matches("default")) Error(
					"case or default expected", in.token
				); comment("<DEFAULT>");
			}
			parseBlock(context.passSwitch(next, end), false);
			if(notdef) PostLabel(next, "<End CASE>");
			else PostLabel(next, "<End DEFAULT>");
		} while(true);
		PostLabel(end,"<End SWITCH>");
	}

	public Result parseAssignment(Context context, variable var) {
		// Assume that the first '=' is already parsed
		if(var.type.dim != 0) Error(
			"You cannot assign a value to an array (you can\n"+
			"assign values to array indexes though)", var.toString()
		); Result result = parseExpression(
			context, var.type, Result.freeList(), false
		); if(in.getCharIf('=')) {
			if(result.address) Error(
				"You cannot assign a value to an address", var.toString()
			); result = parseAssignment(
				context, new variable(result.type, null, result.value)
			);
		} result.loadRegister(this, Result.freeList());
		asmOutln(" LD ("+var.address+"),"+result.value);
		if(var.name != null) {
			if(var.address.indexOf("local_vars+") == 0)
				use(context, context.trail+var.name);
			else use(context, var.name);
		} return result;
	}

	public boolean parseBlock(Context context, boolean If) {
		boolean end = false;
		boolean returned = false;
		boolean bracket = in.getCharIf('{');
		if(!bracket && in.frontChar(in.next())) {
			in.mark(); in.getIdent();
			end = If == in.Matches("then") && If != in.Matches("begin");
			if(end) in.unMark();
			else in.pushBack();
		}
		do {
			do{
				do{}while(in.getCharIf(';'));
				if(in.frontChar(in.next())) {
					in.mark(); if(
						in.getIdent().equals("TRUE") || in.Matches("FALSE") ||
						in.Matches("SFLAG") || in.Matches("NSFLAG") ||
						in.Matches("ZFLAG") || in.Matches("NZFLAG") ||
						in.Matches("HFLAG") || in.Matches("NHFLAG") ||
						in.Matches("PEFLAG") || in.Matches("POSFLAG") ||
						in.Matches("NFLAG") || in.Matches("NNFLAG") ||
						in.Matches("CFLAG") || in.Matches("NCFLAG")
					) continue;
					in.pushBack();
				}
			} while(false);
			in.checkForAsm(this, context);
			if(bracket) {
				if(in.getCharIf('}')) {
					clearUnused(context.previous);
					if(If) {
						if(!in.eof()) {
							if(in.frontChar(in.next())) {
								in.mark();
								if(in.getIdent().equals("else")) {
									in.unMark(); return true;
								} in.pushBack();
							}
						} return false;
					} return returned;
				}
			} in.getIdent();
			in.checkForAsm(this, context);
			if(in.Matches("end")) {
				if(bracket) Error("\'end\' found where '}' was expected");
				if(!end)
					Error("end found where a statement was expected");
				clearUnused(context.previous);
				if(If) return false;
				return returned;
			}
			if(in.Matches("else")) {
				clearUnused(context.previous);
				if(If) return true;
				Error("\"else\" found without matching \"if\"");
			}
			if(returned && loading)
				Warning("[unreachable] statements found after return");
			returned = false;
			if(in.Matches("if")) parseIf(context);
			else if(in.Matches("while") || in.Matches("until")) {
				parseWhile(context, in.Matches("until"));
			}
			else if(in.Matches("do")) parseDo(context);
			else if(in.Matches("loop")) parseLoop(context);
			else if(in.Matches("for")) parseFor(context);
			else if(in.Matches("switch")) parseSwitch(context);
			else if(in.Matches("goto")) parseGoto(context);
			else if(in.Matches("label")) {
				in.getIdent();
				if(loading) {
					if(utils.find(func.labels, in.token) >= 0) Error(
						"Duplicate \""+in.token+"\" label in function "+
						func.ret.name
					); func.labels = utils.add(func.labels, in.token);
				} asmOutln("Lbl_"+func.ret.name+'_'+in.token+':');
				in.MatchChar(';');
			} else if(in.Matches("continue")) {
				if(context.cont < 0) Error("No loop to continue");
				jump(context.cont, "continue"); in.MatchChar(';');
			} else if(in.Matches("break")) {
				if(context.brk < 0) Error("No loop to break from");
				jump(context.brk, "break"); in.MatchChar(';');
			} else if(in.Matches("next")) {
				if(context.next < 0) Error("No switch to continue");
				jump(context.next, "next"); in.MatchChar(';');
			} else if(in.Matches("exit")) {
				if(context.exit < 0) Error("No switch to exit from");
				jump(context.exit, "exit"); in.MatchChar(';');
			} else if(in.Matches("return")) {
				asmOutln(" RET"); in.MatchChar(';'); returned = true;
			} else {
				variable var = null;
				int[] typeof = null;
				String ident = in.token;
				if(Type.valid(ident, false)) {
					Type type = in.getType(false);
					do {
						ident = in.getIdent();
						typeof = typeOf(context, ident);
						switch(typeof[0]) {
						  case 0: Error(
								"\""+ident +
								"\" is already defined as a function",
								funcs[typeof[1]].toString()
							);
						  case 1: Error(
								"\""+ident +
								"\" is already defined as a global variable",
								globals[typeof[1]].toString()
							);
						  case 2: Error(
								"\""+ident +
								"\" is already defined as a local variable",
								context.varAt(typeof[1]).toString()
							);
						  case 3: Error("You cannot name a variable \"this\"");
						}
						String address = "$0000";
						if(!loading) address =
							"local_vars+"+newAddress(context,type.size());
						var =  new variable(type, ident, address);
						context.locals = utils.add(context.locals, var);
						if(in.next() == '(') Error(
							"Syntax Error", type.toString()+ident+'('
						); if(in.getCharIf('=')) parseAssignment(context, var);
					} while(in.getCharIf(','));
				} else {
					typeof = typeOf(context, ident);
					switch(typeof[0]) {
					  case 0:
						in.MatchChar('(');
						use(context, funcs[typeof[1]].ret.name);
						parseFunctionCall(
							context, funcs[typeof[1]], Result.freeList()
						); break;
					  case 1:
						var = globals[typeof[1]];
						use(context, var.name);
						break;
					  case 2:
						var = context.varAt(typeof[1]);
						use(context, context.nameAt(typeof[1]));
						break;
					  case 3:
						var = func.ret;
						noReturn = false;
						break;
					  default: Error("Undefined identifier: "+ident);
					}
					if(var != null) {
						if(!in.getCharIf('=')) Error(
							"Syntax Error: "+ident+' '+in.getToken(true)
						); parseAssignment(context, var);
					}
				} in.MatchChar(';');
			}
		} while(bracket || end);
		clearUnused(context.previous);
		if(If) {
			if(!in.eof()) {
				if(in.frontChar(in.next())) {
					in.mark();
					if(in.getIdent().equals("else")) {
						in.unMark(); return true;
					} in.pushBack();
				}
			} return false;
		} return returned;
	}

	public boolean available(Context context, int address) {
		if(memory[address].length == 0) return true;
		int index, size, i;
		for(i=0; i<memory[address].length; i++) {
			index = utils.find(funcs, memory[address][i]);
			if(index >= 0) {
				if(
					funcs[index] != func && (
						utils.find(funcs[index].used, func.ret.name) >= 0 ||
						utils.find(func.used, funcs[index].ret.name) >= 0
					)
				) return false;
			}
		}
		if(utils.find(memory[address], func.ret.name) < 0) return true;
		size = context.length();
		for(i=0; i<size; i++) {
			if(
				(context.varAt(i).getAddress() <= address) && (
					context.varAt(i).getAddress() >
					address-context.varAt(i).type.size()
				)
			) return false;
		} return true;
	}

	public int newAddress(Context context, int bytes) {
		int j, index;
		boolean pass;
		for(int i=0; i<memory.length-bytes+1; i++) {
			pass = true;
			for(j=0; j<bytes && pass; j++)
				pass = available(context, i+j);
			if(pass) {
				for(j=0; j<bytes && pass; j++) {
					if(utils.find(memory[i+j], func.ret.name) < 0)
						memory[i+j] = utils.add(memory[i+j], func.ret.name);
				} return i;
			}
		} Error("Ran out of usable runtime memory (1000 bytes wasn't enough?)");
		return -1;
	}

	private int nextValue() {
		String str;
		int size = 0;
		if(in.next() == '\"') {
			str = in.getString(true);
			size = str.length()-2;
			char ch;
			for(int i=1; i<str.length()-1; i++) {
				ch = str.charAt(i);
				if(ch=='\n' || ch=='\t' || ch=='\0') {
					if(i < str.length()-2) {
						str = str.substring(0,i)+"\","+(byte)ch +
							",\""+str.substring(i+1);
						i += 4;
					} else {
						str = str.substring(0,i)+"\","+(byte)ch;
						i += 2;
					}
				}
			}
		}
		else {
			str = null;
			int value = in.getWholeNum();
			do {
				size++;
				if(str == null) str = Integer.toString(value%256);
				else str += ","+(value%256); // z80 is LITTLE-ENDIAN
				value /= 256;
			} while(value > 0);
		}
		in.token = str;
		return size;
	}

	public String initGlobal(Type type) {
		if(type.dim == 0) {
			int size = nextValue();
			if(size > type.bytes) Error(
				"Value too big for datatype ("+type.toString()+')',
				in.token
			);
			return "0,".substring(0,2*(type.bytes-size))+in.token;
		}
		int i = 0;
		String init = "";
		in.MatchChar('{');
		while(!in.getCharIf('}')) {
			if(i > 0) {
				in.MatchChar(',');
				init += ',';
			}
			i += nextValue();
			init += in.token;
		}
		if(i != type.dim) Error(
			"Incorrect amount of bytes in array/n(" +
			type.dim+"bytes expected, "+i+" bytes found)",
			"{"+init+'}'
		);
		return init;
	}

	public String parseSource() {
		if(in.eof()) {
			Warning("source has is nothing in it");
			return null;
		}
		Type type;
		int index;
		String str;
		int[] sizes;
		int[] typeof;
		variable var;
		variable[] args;
		Context context;
		String header = "";
		while(!in.eof()) {
			in.checkForAsm(null, null);
			if(in.getIf('#')) {
				if(in.getIdent().equals("asminc")) {
					str = in.getString(false);
					if(utils.find(asmincs, str) == -1)
						asmincs = utils.add(asmincs, str);
				}
				else if(in.Matches("header"))
					header += in.getString(false)+'\n';
				else Error("Invalid command here: #"+in.token);
			}
			else { // Do NOT "use" anything here; things get used in functions
				in.getIdent();
				if(!Type.valid(in.token, true)) Error(
					"Global variable or function expected", in.token
				);
				var = in.getVariable(false);
				typeof = typeOf(new Context(), var.name);
				if(typeof[0] == 0) {
					if(funcs[typeof[1]].code != null) Error(
						var.name+" is already defined as a function"
					);
				}
				if(typeof[0] > 0) Error(
					var.name+" is already defined as a global variable"
				);
				if(in.getCharIf('(')) {
					args = new variable[0];
					if(!in.getCharIf(')')) {
						do {
							type = in.getType(true);
							if(in.next() == ',') str = null;
							else str = in.getIdent();
							args = utils.add(
								args, new variable(type, str, "$0000")
							);
						} while(in.getChar() == ',');
					}
					if(!in.Matches(")"))
						Error("')' expected (next char is \'"+in.token+'\'');
					func = new function(var, args);
					index = utils.find(funcs, var.name);
					if(in.getCharIf(';')) {
						if(index >= 0) Error(
							var.name+" is already defined as a function"
						); func.code = null;
						funcs = utils.add(funcs, func);
					}
					else {
						if(index >= 0) {
							if(funcs[index].code != null) Error(
								var.name+" is already defined as a function"
							);
							if(funcs[index].args.length != args.length) Error(
								"Function "+var.name +
								" does not match its header"
							);
							for(int i=0; i<args.length; i++) {
								if(!args[i].type.equals(
									funcs[index].args[i].type
								)) Error(
									"Function "+var.name +
									" does not match its header"
								);
								if(args[i].name == null) {
									if(funcs[index].args[i].name == null) {
										Error(
											"no name given for argument #"+i+
											" in function "+var.name
										);
									} args[i] = funcs[index].args[i];
								}
							} funcs[index] = func;
						}
						else {
							for(int i=0; i<args.length; i++) {
								if(args[i].name == null) Error(
									"no name given for argument #"+i+
									" in function "+var.name
								);
							} funcs = utils.add(funcs, func);
						}
						in.clear();
						noReturn = true;
						context = new Context();
						context.locals = args;
						parseBlock(context, false);
						if(noReturn && (func.ret.type.bytes > 0)) {
							Warning(
								"No return value assigned in function "+func+
								"\n(expect it to return a garbage value)"
							);
						}
						for(int j=0; j<func.args.length; j++) {
							func.args[j].address = "local_vars+"+newAddress(
								context, func.args[j].type.size()
							);
						}
						if(func.ret.type.bytes > 0) {
							func.ret.address = "local_vars+"+newAddress(
								context, func.ret.type.size()
							);
						}
						func.code = in.output;
						in.clear();
					} func = null;
				}
				else {
					boolean repeat;
					do {
						if(var.type.bytes == 0)
							Error("Variables cannot be void", var.toString());
						if(!in.eof()) {
							if(in.getCharIf('@')) {
								str = null; var.address = in.getString(false);
							} else {
								var.address = "global_"+var.name;
								if(in.getCharIf('=')) str = initGlobal(var.type);
								else str = "0,0".substring(
									0, 2*(var.type.size())-1
								);
							}
						} else {
							var.address = "global_"+var.name;
							str = "0,0".substring(0,1+2*(var.type.dim-1));
						} inits = utils.add(inits, str);
						globals = utils.add(globals, var);
						repeat = in.getCharIf(',');
						if(repeat) {
							var = new variable(var.type, null, "$0000");
							var.name = in.getIdent();
							typeof = typeOf(new Context(), var.name);
							if(typeof[0] == 0) {
								if(funcs[typeof[1]].code != null) Error(
									var.name+" is already defined as a function"
								);
							}
							if(typeof[0] > 0) Error(
								var.name+
								" is already defined as a global variable"
							);
						}
					} while(repeat);
					in.MatchChar(';');
				}
			}
		} return header;
	}

	private boolean[][] removeUnused(boolean[][] used, int f) {
		int index;
		for(int i=0; i<funcs[f].used.length; i++) {
			index = utils.find(globals, funcs[f].used[i]);
			if(index >= 0) used[1][index] = true;
			else {
				index = utils.find(funcs, funcs[f].used[i]);
				if(index >= 0) {
					if(!used[0][index]) {
						used[0][index] = true;
						used = removeUnused(used, index);
					}
				}
			}
		} return used;
	}

	public void removeUnused() {
		boolean[][] used = new boolean[][] {
			new boolean[funcs.length],
			new boolean[globals.length]
		};
		int index = utils.find(funcs, "main");
		if(index < 0) Error("No main function was found");
		used[0][index] = true;
		used = removeUnused(used, index);
		boolean[] temp;
		function[] tfunc;
		for(int i=0; i<used[0].length; i++) {
			if(!used[0][i]) {
				temp = new boolean[used[0].length-1];
				tfunc = new function[funcs.length-1];
				for(int j=0; j<i; j++) {
					temp[j] = used[0][j];
					tfunc[j] = funcs[j];
				}
				for(int j=i; j<used[0].length-1; j++) {
					temp[j] = used[0][j+1];
					tfunc[j] = funcs[j+1];
				}
				used[0] = temp;
				funcs = tfunc;
				i--;
			}
		} String[] tinits;
		variable[] tglob;
		for(int i=0; i<used[1].length; i++) {
			if(!used[1][i]) {
				temp = new boolean[used[1].length-1];
				tinits = new String[inits.length-1];
				tglob = new variable[globals.length-1];
				for(int j=0; j<i; j++) {
					temp[j] = used[1][j];
					tinits[j] = inits[j];
					tglob[j] = globals[j];
				}
				for(int j=i; j<used[1].length-1; j++) {
					temp[j] = used[1][j+1];
					tinits[j] = inits[j+1];
					tglob[j] = globals[j+1];
				} used[1] = temp;
				inits = tinits;
				globals = tglob;
				i--;
			}
		}
	}

	public void compile() {
		Context context;
		String text, arg;
		boolean none = true;
		for(current=0; current<asmUsed.length; current++) {
			usecphlde = false;
			if(!asmUsed[current]) continue;
			show(current, true);
			none = false;
			Labels = 0;
			inits = new String[0];
			asmincs = new String[0];
			funcs = new function[0];
			globals = new variable[0];
			memory = new String[1000][0];
			in = new Parser(code[current], this);
			try {
				loading = true;
				String asm = parseSource();
				loading = false;
				removeUnused();
				for(int i=0; i<asmincs.length; i++)
					asmOutln("#include \""+asmincs[i]+'\"');
				if(asm.length() > 0) asmOut(asm);
				int size = utils.find(funcs, "main");
				if(size < 0) Error("No main function was found");
				if(size != 0) {
					func = funcs[0];
					funcs[0] = funcs[size];
					funcs[size] = func;
				}
				if(funcs[0].ret.type.bytes > 0) Error(
					"The main function must have a void return type"
				);
				for(int i=0; i<funcs.length; i++) {
					func = funcs[i];
					if(func.code == null) Error(
						"Unused function header:\n"+func
					); context = new Context();
					context.locals = func.args;
					in = new Parser(func.code, this);
					asmOutln("\nfunc_"+func.ret.name+':');
					if(!parseBlock(context, false)) asmOutln(" RET");
				}
				if(usecphlde) {
					asmOutln("\ncphlde:");
					asmOutln(" push hl");
					asmOutln(" or a");
					asmOutln(" sbc hl,de");
					asmOutln(" pop hl");
					asmOutln(" ret");
				}
				for(int i=0; i<globals.length; i++) {
					if(inits[i] == null) continue;
					asmOut("\n"+globals[i].address+':');
					comment(globals[i].toString());
					asmOutln(" .DB "+inits[i]);
				}
				if(memory[0].length > 0) {
					asmOutln("\nlocal_vars:");
					for(int i=0; i<memory.length; i+=25) {
						asmOut(" .DB ");
						for(int j=0; j<25; j++) {
							if(memory[i+j].length == 0) {
								loading = true; break;
							}
							if(j!=0) asmOut(",");
							asmOut("0");
						}
						if(loading) {
							loading = false;
							break;
						}
						asmOutln();
					}
				}
				asmOutln("\n.END\n.END");
			}
			catch(CompileError c) {
				if(asmOutput[current].getText().length() > 0) {
					asmOutput[current].setText(
						"; ERRORS WERE FOUND IN THIS " +
						"PROGRAM AND IT IS NOT SAFE\n" +
						"; TO RUN, SO DO NOT ATTEMPT " +
						"COMPILE AND RUN THIS PROGRAM.\n\n" +
						asmOutput[current].getText()+"\n\n" +
						"; COMPILATION STOPPED HERE.\n" +
						"; THIS PROGRAM IS INCOMPLETE " +
						"BECAUSE THE SOURCE HAD ERRORS"
					);
				}
				String stop = "Stop compilating";
				if(
					javax.swing.JOptionPane.OK_OPTION ==
					javax.swing.JOptionPane.showOptionDialog(
						root,
						"Compilation for the "+asmNames[current] +
						" has been\nstopped because errors were found",
						"SquirrelBox",
						javax.swing.JOptionPane.OK_CANCEL_OPTION,
						javax.swing.JOptionPane.WARNING_MESSAGE,
						icon, new String[]{stop,"Next platform"}, stop
					)
				) break;
			}
		}
		if(none) Message(
			"No platform was defined, so there was nothing to\n" +
			"compile. You must define at least one platform in\n" +
			"your code by using the #define command with one or\n" +
			"more of the following constants:\n\n" +

			"TI82, TI83, TI83PLUS, TI83PLUSSE,\n" +
			"TI84PLUS, TI84PLUSSE, TI85, TI86\n\n" +

			"(i.e. to compile for the TI-83+: #define TI83PLUS)"
		);
		Message(
			"Compilation complete. Save any output code generated\n" +
			"as z80 assembly source files, and then use an assembly\n" +
			"compiler to compile them. DO NOT ATTEMPT TO COMPILE\n" +
			"AND RUN PROGRAMS THAT HAD ERRORS IN THEM, they\n" +
			"are NOT SAFE and will crash your calc (and might destroy\n" +
			"other data that you have on there) if you do it anyway"
		); in = null;
	}

	public static void main(String[] args) {
		javax.swing.JFrame win = new javax.swing.JFrame(
			"SquirrelBox"
		);
		win.setDefaultCloseOperation(
			javax.swing.WindowConstants.DO_NOTHING_ON_CLOSE
		);
		final HLz80compiler compiler = new HLz80compiler();
		compiler.init(win.getRootPane(), true);
		win.setIconImage(compiler.icon.getImage());
		win.addWindowListener(new java.awt.event.WindowAdapter() {
			public void windowClosing(java.awt.event.WindowEvent w) {
				compiler.exit();
			}
			public void windowClosed(java.awt.event.WindowEvent w) {
				compiler.exit();
			}
		});
		win.setBounds(75,50,650,550);
		win.setVisible(true);
		compiler.Message(
"(the) SquirrelBox compiler v1.0.\n" +
"Copyright (C) 2005 Dan Cook\n\n" +

"This program is free software; you can redistribute it and/or\n" +
"modify it under the terms of the GNU General Public License\n" +
"as published by the Free Software Foundation; either version 2\n" +
"of the License, or (at your option) any later version.\n\n" +

"This program is distributed in the hope that it will be useful,\n" +
"but WITHOUT ANY WARRANTY; without even the implied warranty of\n" +
"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the\n" +
"GNU General Public License for more details.\n\n" +

"You should have received a copy of the GNU General Public License\n" +
"along with this program; if not, write to the Free Software\n" +
"Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA."
		);
	}
}